////////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2001-2025 Alexey Kuryakin daqgroup@mail.ru under MIT license //
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// This file is part of the CRW-DAQ project by DaqGroup - component CRWLIB.   //
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Purpose:                                                                   //
// DataBase API constants. Derived from ADO etc.                              //
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// History:                                                                   //
// 20221031 - 1st release                                                     //
// 20221119 - AdoFieldTypeToString                                            //
// 20230826 - Modified for FPC (A.K.)                                         //
// 20250625 - DbEventTypeToString, DbFieldTypeCodeToString, etc               //
////////////////////////////////////////////////////////////////////////////////

unit _crw_dbcon; // DataBase Constants

{$I _crw_sysdef.inc}

{$I _crw_sysmode.inc}

{$WARN 5023 off : Unit "$1" not used in $2}

interface

uses
 //////////////////////////////////////////////////////
 {$I _crw_uses_first.inc} // NB: MUST BE FIRST USES !!!
 //////////////////////////////////////////////////////
 sysutils, classes, math, db, sqltypes, sqldb,
 _crw_alloc, _crw_ef, _crw_rtc, _crw_str;

const //////////////////// ADO DB constants; ADO = ActiveX Data Objects
 db_engine_code_ado      = 1;       // ADO engine code
 db_engine_name_ado      = 'ADO';   // ADO engine name

const //////////////////// SQLDB constants; SQLDB = default Lazarus DB engine
 db_engine_code_sqldb    = 2;       // SQLDB engine code
 db_engine_name_sqldb    = 'SQLDB'; // SQLDB engine name

////////////////////////////////////////////////////////////////////////////////
// ADO constants derived from AdoInt.pas with next command:
// type "c:\Program Files\Borland\Delphi5\Source\Vcl\adoint.pas" ^
//   | unix grep -i -e '^\s*ad[_a-zA-Z][_a-zA-Z0-9]*\s*=\s*[\$0-9\-]'  ^
//   | unix gawk '{printf " %%-30s %%s %%-10s\n", $1, $2, $3 }' ^
//   | unix u2d
////////////////////////////////////////////////////////////////////////////////

///////////////////////
// AdoInt.pas constants
///////////////////////
const
 ///////////////////////////////// CursorTypeEnum constants
 adOpenUnspecified              = -1;
 adOpenForwardOnly              = $00000000;
 adOpenKeyset                   = $00000001;
 adOpenDynamic                  = $00000002;
 adOpenStatic                   = $00000003;
 //////////////////////////////// CursorOptionEnum constants
 adHoldRecords                  = $00000100;
 adMovePrevious                 = $00000200;
 adAddNew                       = $01000400;
 adDelete                       = $01000800;
 adUpdate                       = $01008000;
 adBookmark                     = $00002000;
 adApproxPosition               = $00004000;
 adUpdateBatch                  = $00010000;
 adResync                       = $00020000;
 adNotify                       = $00040000;
 adFind                         = $00080000;
 adSeek                         = $00400000;
 adIndex                        = $00800000;
 //////////////////////////////// LockTypeEnum constants
 adLockUnspecified              = -1;
 adLockReadOnly                 = $00000001;
 adLockPessimistic              = $00000002;
 adLockOptimistic               = $00000003;
 adLockBatchOptimistic          = $00000004;
 //////////////////// ExecuteOptionEnum constants
 adOptionUnspecified            = -1;
 adAsyncExecute                 = $00000010;
 adAsyncFetch                   = $00000020;
 adAsyncFetchNonBlocking        = $00000040;
 adExecuteNoRecords             = $00000080;
 //////////////////////////////// ConnectOptionEnum constants
 adConnectUnspecified           = -1;
 adAsyncConnect                 = $00000010;
 //////////////////////////////// ObjectStateEnum constants
 adStateClosed                  = $00000000;
 adStateOpen                    = $00000001;
 adStateConnecting              = $00000002;
 adStateExecuting               = $00000004;
 adStateFetching                = $00000008;
 /////////////////////////////// CursorLocationEnum constants
 adUseNone                      = $00000001;
 adUseServer                    = $00000002;
 adUseClient                    = $00000003;
 adUseClientBatch               = $00000003;
 //////////////////////////////// DataTypeEnum constants
 adEmpty                        = $00000000; // 0   No value
 adTinyInt                      = $00000010; // 16  A 1-byte signed integer.
 adSmallInt                     = $00000002; // 2   A 2-byte signed integer.
 adInteger                      = $00000003; // 3   A 4-byte signed integer.
 adBigInt                       = $00000014; // 20  An 8-byte signed integer.
 adUnsignedTinyInt              = $00000011; // 17  A 1-byte unsigned integer.
 adUnsignedSmallInt             = $00000012; // 18  A 2-byte unsigned integer.
 adUnsignedInt                  = $00000013; // 19  A 4-byte unsigned integer.
 adUnsignedBigInt               = $00000015; // 21  An 8-byte unsigned integer.
 adSingle                       = $00000004; // 4   A single-precision floating-point value.
 adDouble                       = $00000005; // 5   A double-precision floating-point value.
 adCurrency                     = $00000006; // 6   A currency value
 adDecimal                      = $0000000E; // 14  An exact numeric value with a fixed precision and scale.
 adNumeric                      = $00000083; // 131 An exact numeric value with a fixed precision and scale.
 adBoolean                      = $0000000B; // 11  A boolean value.
 adError                        = $0000000A; // 10  A 32-bit error code
 adUserDefined                  = $00000084; // 132 A user-defined variable.
 adVariant                      = $0000000C; // 12  An Automation Variant. Note: Currently not supported by ADO.
 adIDispatch                    = $00000009; // 9   A pointer to an IDispatch interface on a COM object. Note: Currently not supported by ADO.
 adIUnknown                     = $0000000D; // 13  A pointer to an IUnknown interface on a COM object. Note: Currently not supported by ADO.
 adGUID                         = $00000048; // 72  A globally unique identifier (GUID)
 adDate                         = $00000007; // 7   The number of days since December 30, 1899 + the fraction of a day.
 adDBDate                       = $00000085; // 133 A date value (yyyymmdd).
 adDBTime                       = $00000086; // 134 A time value (hhmmss).
 adDBTimeStamp                  = $00000087; // 135 A date/time stamp (yyyymmddhhmmss plus a fraction in billionths).
 adBSTR                         = $00000008; // 8   A null-terminated character string.
 adChar                         = $00000081; // 129 A string value.
 adVarChar                      = $000000C8; // 200 A string value (Parameter object only).
 adLongVarChar                  = $000000C9; // 201 A long string value.
 adWChar                        = $00000082; // 130 A null-terminated Unicode character string.
 adVarWChar                     = $000000CA; // 202 A null-terminated Unicode character string.
 adLongVarWChar                 = $000000CB; // 203 A long null-terminated Unicode string value.
 adBinary                       = $00000080; // 128 A binary value.
 adVarBinary                    = $000000CC; // 204 A binary value (Parameter object only).
 adLongVarBinary                = $000000CD; // 205 A long binary value.
 adChapter                      = $00000088; // 136 A 4-byte chapter value that identifies rows in a child rowset
 adFileTime                     = $00000040; // 64  The number of 100-nanosecond intervals since January 1,1601
 adDBFileTime                   = $00000089; // 137 Database file time
 adPropVariant                  = $0000008A; // 138 An Automation PROPVARIANT.
 adVarNumeric                   = $0000008B; // 139 A numeric value (Parameter object only).
 adArray                        = $00002000; // 0x2000 A flag value combined with another data type constant. Indicates an array of that other data type.
 //////////////////////////////// FieldAttributeEnum constants
 adFldUnspecified               = -1;
 adFldMayDefer                  = $00000002;
 adFldUpdatable                 = $00000004;
 adFldUnknownUpdatable          = $00000008;
 adFldFixed                     = $00000010;
 adFldIsNullable                = $00000020;
 adFldMayBeNull                 = $00000040;
 adFldLong                      = $00000080;
 adFldRowID                     = $00000100;
 adFldRowVersion                = $00000200;
 adFldCacheDeferred             = $00001000;
 adFldNegativeScale             = $00004000;
 adFldKeyColumn                 = $00008000;
 //////////////////////////////// EditModeEnum constants
 adEditNone                     = $00000000;
 adEditInProgress               = $00000001;
 adEditAdd                      = $00000002;
 adEditDelete                   = $00000004;
 //////////////////////////////// RecordStatusEnum constants
 adRecOK                        = $00000000;
 adRecNew                       = $00000001;
 adRecModified                  = $00000002;
 adRecDeleted                   = $00000004;
 adRecUnmodified                = $00000008;
 adRecInvalid                   = $00000010;
 adRecMultipleChanges           = $00000040;
 adRecPendingChanges            = $00000080;
 adRecCanceled                  = $00000100;
 adRecCantRelease               = $00000400;
 adRecConcurrencyViolation      = $00000800;
 adRecIntegrityViolation        = $00001000;
 adRecMaxChangesExceeded        = $00002000;
 adRecObjectOpen                = $00004000;
 adRecOutOfMemory               = $00008000;
 adRecPermissionDenied          = $00010000;
 adRecSchemaViolation           = $00020000;
 adRecDBDeleted                 = $00040000;
 //////////////////////////////// GetRowsOptionEnum constants
 adGetRowsRest                  = -1;
 //////////////////////////////// PositionEnum constants
 adPosUnknown                   = -1;       
 adPosBOF                       = $FFFFFFFE;
 adPosEOF                       = $FFFFFFFD;
 //////////////////////////////// BookmarkEnum constants
 adBookmarkCurrent              = $00000000;
 adBookmarkFirst                = $00000001;
 adBookmarkLast                 = $00000002;
 //////////////////////////////// MarshalOptionsEnum constants
 adMarshalAll                   = $00000000;
 adMarshalModifiedOnly          = $00000001;
 //////////////////////////////// AffectEnum constants
 adAffectCurrent                = $00000001;
 adAffectGroup                  = $00000002;
 adAffectAll                    = $00000003;
 adAffectAllChapters            = $00000004;
 //////////////////////////////// ResyncEnum constants
 adResyncUnderlyingValues       = $00000001;
 adResyncAllValues              = $00000002;
 //////////////////////////////// CompareEnum constants
 adCompareLessThan              = $00000000;
 adCompareEqual                 = $00000001;
 adCompareGreaterThan           = $00000002;
 adCompareNotEqual              = $00000003;
 adCompareNotComparable         = $00000004;
 //////////////////////////////// FilterGroupEnum constants
 adFilterNone                   = $00000000;
 adFilterPendingRecords         = $00000001;
 adFilterAffectedRecords        = $00000002;
 adFilterFetchedRecords         = $00000003;
 adFilterPredicate              = $00000004;
 adFilterConflictingRecords     = $00000005;
 //////////////////////////////// SearchDirectionEnum constants
 adSearchForward                = $00000001;
 adSearchBackward               = -1;
 //////////////////////////////// PersistFormatEnum constants
 adPersistADTG                  = $00000000;
 adPersistXML                   = $00000001;
 //////////////////////////////// StringFormatEnum constants
 adClipString                   = $00000002;
 //////////////////////////////// ConnectPromptEnum constants
 adPromptAlways                 = $00000001;
 adPromptComplete               = $00000002;
 adPromptCompleteRequired       = $00000003;
 adPromptNever                  = $00000004;
 //////////////////////////////// ConnectModeEnum constants
 adModeUnknown                  = $00000000;
 adModeRead                     = $00000001;
 adModeWrite                    = $00000002;
 adModeReadWrite                = $00000003;
 adModeShareDenyRead            = $00000004;
 adModeShareDenyWrite           = $00000008;
 adModeShareExclusive           = $0000000C;
 adModeShareDenyNone            = $00000010;
 //////////////////////////////// IsolationLevelEnum constants
 adXactUnspecified              = -1;
 adXactChaos                    = $00000010;
 adXactReadUncommitted          = $00000100;
 adXactBrowse                   = $00000100;
 adXactCursorStability          = $00001000;
 adXactReadCommitted            = $00001000;
 adXactRepeatableRead           = $00010000;
 adXactSerializable             = $00100000;
 adXactIsolated                 = $00100000;
 //////////////////////////////// XactAttributeEnum constants
 adXactCommitRetaining          = $00020000;
 adXactAbortRetaining           = $00040000;
 adXactAsyncPhaseOne            = $00080000;
 adXactSyncPhaseOne             = $00100000;
 //////////////////////////////// PropertyAttributesEnum constants
 adPropNotSupported             = $00000000;
 adPropRequired                 = $00000001;
 adPropOptional                 = $00000002;
 adPropRead                     = $00000200;
 adPropWrite                    = $00000400;
 //////////////////////////////// ErrorValueEnum constants
 adErrInvalidArgument           = $00000BB9;
 adErrNoCurrentRecord           = $00000BCD;
 adErrIllegalOperation          = $00000C93;
 adErrInTransaction             = $00000CAE;
 adErrFeatureNotAvailable       = $00000CB3;
 adErrItemNotFound              = $00000CC1;
 adErrObjectInCollection        = $00000D27;
 adErrObjectNotSet              = $00000D5C;
 adErrDataConversion            = $00000D5D;
 adErrObjectClosed              = $00000E78;
 adErrObjectOpen                = $00000E79;
 adErrProviderNotFound          = $00000E7A;
 adErrBoundToCommand            = $00000E7B;
 adErrInvalidParamInfo          = $00000E7C;
 adErrInvalidConnection         = $00000E7D;
 adErrNotReentrant              = $00000E7E;
 adErrStillExecuting            = $00000E7F;
 adErrOperationCancelled        = $00000E80;
 adErrStillConnecting           = $00000E81;
 adErrNotExecuting              = $00000E83;
 adErrUnsafeOperation           = $00000E84;
 //////////////////////////////// ParameterAttributesEnum constants
 adParamSigned                  = $00000010;
 adParamNullable                = $00000040;
 adParamLong                    = $00000080;
 //////////////////////////////// ParameterDirectionEnum constants
 adParamUnknown                 = $00000000;
 adParamInput                   = $00000001;
 adParamOutput                  = $00000002;
 adParamInputOutput             = $00000003;
 adParamReturnValue             = $00000004;
 //////////////////////////////// CommandTypeEnum constants
 adCmdUnspecified               = -1;       
 adCmdUnknown                   = $00000008;
 adCmdText                      = $00000001;
 adCmdTable                     = $00000002;
 adCmdStoredProc                = $00000004;
 adCmdFile                      = $00000100;
 adCmdTableDirect               = $00000200;
 //////////////////////////////// EventStatusEnum constants
 adStatusOK                     = $00000001;
 adStatusErrorsOccurred         = $00000002;
 adStatusCantDeny               = $00000003;
 adStatusCancel                 = $00000004;
 adStatusUnwantedEvent          = $00000005;
 //////////////////////////////// EventReasonEnum constants
 adRsnAddNew                    = $00000001;
 adRsnDelete                    = $00000002;
 adRsnUpdate                    = $00000003;
 adRsnUndoUpdate                = $00000004;
 adRsnUndoAddNew                = $00000005;
 adRsnUndoDelete                = $00000006;
 adRsnRequery                   = $00000007;
 adRsnResynch                   = $00000008;
 adRsnClose                     = $00000009;
 adRsnMove                      = $0000000A;
 adRsnFirstChange               = $0000000B;
 adRsnMoveFirst                 = $0000000C;
 adRsnMoveNext                  = $0000000D;
 adRsnMovePrevious              = $0000000E;
 adRsnMoveLast                  = $0000000F;
 //////////////////////////////// SchemaEnum constants
 adSchemaProviderSpecific       = -1;       
 adSchemaAsserts                = $00000000;
 adSchemaCatalogs               = $00000001;
 adSchemaCharacterSets          = $00000002;
 adSchemaCollations             = $00000003;
 adSchemaColumns                = $00000004;
 adSchemaCheckConstraints       = $00000005;
 adSchemaConstraintColumnUsage  = $00000006;
 adSchemaConstraintTableUsage   = $00000007;
 adSchemaKeyColumnUsage         = $00000008;
 adSchemaReferentialConstraints = $00000009;
 adSchemaTableConstraints       = $0000000A;
 adSchemaColumnsDomainUsage     = $0000000B;
 adSchemaIndexes                = $0000000C;
 adSchemaColumnPrivileges       = $0000000D;
 adSchemaTablePrivileges        = $0000000E;
 adSchemaUsagePrivileges        = $0000000F;
 adSchemaProcedures             = $00000010;
 adSchemaSchemata               = $00000011;
 adSchemaSQLLanguages           = $00000012;
 adSchemaStatistics             = $00000013;
 adSchemaTables                 = $00000014;
 adSchemaTranslations           = $00000015;
 adSchemaProviderTypes          = $00000016;
 adSchemaViews                  = $00000017;
 adSchemaViewColumnUsage        = $00000018;
 adSchemaViewTableUsage         = $00000019;
 adSchemaProcedureParameters    = $0000001A;
 adSchemaForeignKeys            = $0000001B;
 adSchemaPrimaryKeys            = $0000001C;
 adSchemaProcedureColumns       = $0000001D;
 adSchemaDBInfoKeywords         = $0000001E;
 adSchemaDBInfoLiterals         = $0000001F;
 adSchemaCubes                  = $00000020;
 adSchemaDimensions             = $00000021;
 adSchemaHierarchies            = $00000022;
 adSchemaLevels                 = $00000023;
 adSchemaMeasures               = $00000024;
 adSchemaProperties             = $00000025;
 adSchemaMembers                = $00000026;
 adSchemaTrustees               = $00000027;
 //////////////////////////////// SeekEnum constants
 adSeekFirstEQ                  = $00000001;
 adSeekLastEQ                   = $00000002;
 adSeekAfterEQ                  = $00000004;
 adSeekAfter                    = $00000008;
 adSeekBeforeEQ                 = $00000010;
 adSeekBefore                   = $00000020;
 //////////////////////////////// ADCPROP_UPDATECRITERIA_ENUM constants
 adCriteriaKey                  = $00000000;
 adCriteriaAllCols              = $00000001;
 adCriteriaUpdCols              = $00000002;
 adCriteriaTimeStamp            = $00000003;
 //////////////////////////////// ADCPROP_ASYNCTHREADPRIORITY_ENUM constants
 adPriorityLowest               = $00000001;
 adPriorityBelowNormal          = $00000002;
 adPriorityNormal               = $00000003;
 adPriorityAboveNormal          = $00000004;
 adPriorityHighest              = $00000005;
 //////////////////////////////// CEResyncEnum constants
 adResyncNone                   = $00000000;
 adResyncAutoIncrement          = $00000001;
 adResyncConflicts              = $00000002;
 adResyncUpdates                = $00000004;
 adResyncInserts                = $00000008;
 adResyncAll                    = $0000000F;
 //////////////////////////////// ADCPROP_AUTORECALC_ENUM constants
 adRecalcUpFront                = $00000000;
 adRecalcAlways                 = $00000001;

const // ADO data types
 adScalarFieldTypes = [adEmpty,adTinyInt,adSmallInt,adInteger,adBigInt,
 adUnsignedTinyInt,adUnsignedSmallInt,adUnsignedInt, adUnsignedBigInt,
 adSingle,adDouble,adCurrency,adDecimal,adNumeric,adBoolean,adError,
 adUserDefined,adVariant,adIDispatch,adIUnknown,adGUID,adDate,
 adDBDate,adDBTime,adDBTimeStamp,adBSTR,adChar,adVarChar,adLongVarChar,
 adWChar,adVarWChar,adLongVarWChar,adBinary,adVarBinary,adLongVarBinary,
 adChapter,adFileTime,adDBFileTime,adPropVariant,adVarNumeric];

 //////////////////////////////////////////////////////////////////////
 // TDbConverter encapsulate all DataBase related conversion functions.
 // The DbCon is the only one instance of TDbConverter.
 //////////////////////////////////////////////////////////////////////
type
 TDbConverter = class(TMasterObject)
 private
  myDbFieldTypeMap:TStringList;
 protected
  function DbFieldTypeMap:TStringList;
 public
  constructor Create;
  destructor  Destroy; override;
  procedure   AfterConstruction; override;
  procedure   BeforeDestruction; override;
 public
  class var UsesSystemDbFieldTypeNames:Boolean;
 public // Convert DB event type to string like detExecute.
  function EventTypeToString(EventType:TDbEventType):LongString;
 public // Convert dataset state to string like dsInactive.
  function DataSetStateToString(const State:TDataSetState):LongString;
 public // Convert DB schema type to string like stTables.
  function SchemaTypeToString(SchemaType:TSchemaType):LongString;
 public // Convert SQL statment type to string like stSelect.
  function StatementTypeToString(st:TStatementType):LongString;
 public // Convert connection options to string.
  function ConnOptionToString(Opt:TConnOption):LongString;
  function ConnOptionsToString(const Opts:TConnOptions; Sep:Char=','):LongString;
 public // Convert SQL connection options to string.
  function SQLConnectionOptionToString(opt:TSQLConnectionOption):LongString;
  function SQLConnectionOptionsToString(opts:TSQLConnectionOptions; Sep:Char=','):LongString;
 public // Convert ADO field type to string like adInteger.
  function AdoFieldTypeToString(FieldType:Integer):LongString;
 public // Convert SQLDB field type to string like ftInteger.
  function SqlDbFieldTypeToString(FieldType:TFieldType):LongString;
 public // Convert string to DB field type integer code or return default.
        // String must to have prefix like adInteger (ADO), ftInteger (SQDB).
  function StringToFieldTypeCode(const S:LongString; Def:Integer=-1):Integer;
 public // Convert DB field type integer code to string.
        // Return string like adInteger or ftInteger depending on Engine (ADO,SQLDB).
  function FieldTypeCodeToString(Code:Integer; Engine:Integer):LongString;
 public // Convert integer code to TFieldType (with range checking).
  function CodeToFieldType(Code:Integer; Def:TFieldType=ftUnknown):TFieldType;
 public // DateTime converters
  function IsDateTimeFieldTypeCode(typ,Engine:Integer):Boolean;
  function AdoDateTimeToMs(tm:Double; typ:Integer):Double;
  function SqlDbDateTimeToMs(tm:Double; typ:Integer):Double;
  function DbDateTimeToMs(tm:Double; typ,Engine:Integer):Double;
 end;

function DbCon:TDbConverter; // The only one DB converter.

implementation

var
 TheDbCon:TDbConverter=nil;

function DbCon:TDbConverter;
begin
 if not Assigned(TheDbCon) then begin
  TheDbCon:=TDbConverter.Create;
  TheDbCon.Master:=@TheDbCon;
 end;
 Result:=TheDbCon;
end;

procedure Kill(var TheObject:TDbConverter); overload;
begin
 try
  FreeAndNil(TheObject);
 except
  on E:Exception do BugReport(E,nil,'Kill');
 end;
end;

constructor TDbConverter.Create;
begin
 inherited Create;
end;

destructor TDbConverter.Destroy;
begin
 inherited Destroy;
end;

procedure TDbConverter.AfterConstruction;
begin
 inherited AfterConstruction;
 DbFieldTypeMap;
end;

procedure TDbConverter.BeforeDestruction;
begin
 Kill(myDbFieldTypeMap);
 inherited BeforeDestruction;
end;

function TDbConverter.EventTypeToString(EventType:TDbEventType):LongString;
begin
 Result:='';
 case EventType of
  detCustom     : Result:='detCustom';       // Custom event message
  detPrepare    : Result:='detPrepare';      // SQL prepare message
  detExecute    : Result:='detExecute';      // SQLExecute message
  detFetch      : Result:='detFetch';        // Fetch data message
  detCommit     : Result:='detCommit';       // Transaction Commit message
  detRollBack   : Result:='detRollBack';     // Transaction rollback message
  detParamValue : Result:='detParamValue';   // Parameter name and value message
  detActualSQL  : Result:='detActualSQL';    // Actual SQL as sent to engine message
 end;
end;

function TDbConverter.DataSetStateToString(const State:TDataSetState):LongString;
begin
 Result:='';
 case State of
  dsInactive      : Result:='dsInactive';     // The dataset is not active. No data is available.
  dsBrowse        : Result:='dsBrowse';       // The dataset is active, and the cursor can be used to navigate the data.
  dsEdit          : Result:='dsEdit';         // The dataset is in editing mode: the current record can be modified.
  dsInsert        : Result:='dsInsert';       // The dataset is in insert mode: the current record is a new record which can be edited.
  dsSetKey        : Result:='dsSetKey';       // The dataset is calculating the primary key.
  dsCalcFields    : Result:='dsCalcFields';   // The dataset is calculating it's calculated fields.
  dsFilter        : Result:='dsFilter';       // The dataset is filtering records.
  dsNewValue      : Result:='dsNewValue';     // The dataset is showing the new values of a record.
  dsOldValue      : Result:='dsOldValue';     // The dataset is showing the old values of a record.
  dsCurValue      : Result:='dsCurValue';     // The dataset is showing the current values of a record.
  dsBlockRead     : Result:='dsBlockRead';    // The dataset is open, but no events are transferred to datasources.
  dsInternalCalc  : Result:='dsInternalCalc'; // The dataset is calculating it's internally calculated fields.
  dsOpening       : Result:='dsOpening';      // The dataset is currently opening, but is not yet completely open.
  dsRefreshFields : Result:='dsOpening';      // The dataset is refreshing field values from server after update.
 end;
end;

function TDbConverter.SchemaTypeToString(SchemaType:TSchemaType):LongString;
begin
 Result:='';
 case SchemaType of
  stNoSchema        : Result:='stNoSchema';        // No schema
  stTables          : Result:='stTables';          // User Tables in database
  stSysTables       : Result:='stSysTables';       // System tables in database
  stProcedures      : Result:='stProcedures';      // Stored procedures in database
  stColumns         : Result:='stColumns';         // Columns in a table
  stProcedureParams : Result:='stProcedureParams'; // Parameters for a stored procedure
  stIndexes         : Result:='stIndexes';         // Indexes for a table
  stPackages        : Result:='stPackages';        // Packages (for databases that support them)
  stSchemata        : Result:='stSchemata';        // List of schemas in database(s) (for databases that support them)
  stSequences       : Result:='stSequences';       // Sequences (for databases that support them)
 end;
end;

function TDbConverter.StatementTypeToString(st:TStatementType):LongString;
begin
 Result:='';
 case st of
  stUnknown       : Result:='stUnknown';       // The statement type could not be detected.
  stSelect        : Result:='stSelect';        // The statement is a SQL SELECT statement
  stInsert        : Result:='stInsert';        // The statement is a SQL INSERT statement
  stUpdate        : Result:='stUpdate';        // The statement is a SQL UPDATE statement
  stDelete        : Result:='stDelete';        // The statement is a SQL DELETE statement
  stDDL           : Result:='stDDL';           // The statement is a SQL DDL (Data Definition Language) statement
  stGetSegment    : Result:='stGetSegment';    // The statement is a SQL get segment statement
  stPutSegment    : Result:='stPutSegment';    // The statement is a SQL put segment statement
  stExecProcedure : Result:='stExecProcedure'; // The statement executes a stored procedure
  stStartTrans    : Result:='stStartTrans';    // The statement starts a transaction
  stCommit        : Result:='stCommit';        // The statement commits a transaction
  stRollback      : Result:='stRollback';      // The statement rolls back a transaction
  stSelectForUpd  : Result:='stSelectForUpd';  // The statement selects data for update
 end;
end;

function TDbConverter.ConnOptionToString(Opt:TConnOption):LongString;
begin
 Result:='';
 case Opt of
  sqSupportParams            : Result:='sqSupportParams';            // The connection type has native support for parameters.
  sqSupportEmptyDatabaseName : Result:='sqSupportEmptyDatabaseName'; // Does the connection allow empty database names ?
  sqEscapeSlash              : Result:='sqEscapeSlash';              // Escapes in string literals are done with backslash characters.
  sqEscapeRepeat             : Result:='sqEscapeRepeat';             // Escapes in string literals are done by repeating the character.
  sqImplicitTransaction      : Result:='sqImplicitTransaction';      // Does the connection support implicit transaction management
  sqLastInsertID             : Result:='sqLastInsertID';             // Does the connection support getting the ID for the last insert operation.
  sqSupportReturning         : Result:='sqSupportReturning';         // The connection type supports INSERT/UPDATE with RETURNING clause
  sqSequences                : Result:='sqSequences';                // Are sequences supported.
 end;
end;

function TDbConverter.ConnOptionsToString(const Opts:TConnOptions; Sep:Char=','):LongString;
var opt:TConnOption;
begin
 Result:='';
 for opt:=Low(TConnOption) to High(TConnOption) do
 if (opt in Opts) then Result:=Result+ConnOptionToString(opt)+Sep;
 if (Result<>'') then Result:=TrimRightChars(Result,[Sep]);
end;

function TDbConverter.SQLConnectionOptionToString(opt:TSQLConnectionOption):LongString;
begin
 case opt of
  scoExplicitConnect                : Result:='scoExplicitConnect';                // Require explicit connection to the database (default is implicit)
  scoApplyUpdatesChecksRowsAffected : Result:='scoApplyUpdatesChecksRowsAffected'; // ApplyUpdates will check that the RowsAffected is 1 after an update.
 end;
end;

function TDbConverter.SQLConnectionOptionsToString(opts:TSQLConnectionOptions; Sep:Char=','):LongString;
var opt:TSQLConnectionOption;
begin
 Result:='';
 for opt:=Low(TSQLConnectionOption) to High(TSQLConnectionOption) do
 if (opt in Opts) then Result:=Result+SQLConnectionOptionToString(opt)+Sep;
 if (Result<>'') then Result:=TrimRightChars(Result,[Sep]);
end;

 // Converts an ADO native types into string related.
function TDbConverter.AdoFieldTypeToString(FieldType:Integer):LongString;
begin
 Result:='';
 case (FieldType and not adArray) of
  adChar             : Result := 'adChar';
  adVarChar          : Result := 'adVarChar';
  adBSTR             : Result := 'adBSTR';
  adWChar            : Result := 'adWChar';
  adVarWChar         : Result := 'adVarWChar';
  adBoolean          : Result := 'adBoolean';
  adTinyInt          : Result := 'adTinyInt';
  adUnsignedTinyInt  : Result := 'adUnsignedTinyInt';
  adSmallInt         : Result := 'adSmallInt';
  adUnsignedSmallInt : Result := 'adUnsignedSmallInt';
  adInteger          : Result := 'adInteger';
  adUnsignedInt      : Result := 'adUnsignedInt';
  adBigInt           : Result := 'adBigInt';
  adUnsignedBigInt   : Result := 'adUnsignedBigInt';
  adSingle           : Result := 'adSingle';
  adDouble           : Result := 'adDouble';
  adDecimal          : Result := 'adDecimal';
  adNumeric          : Result := 'adNumeric';
  adVarNumeric       : Result := 'adVarNumeric';
  adCurrency         : Result := 'adCurrency';
  adDBDate           : Result := 'adDBDate';
  adDBTime           : Result := 'adDBTime';
  adDate             : Result := 'adDate';
  adDBTimeStamp      : Result := 'adDBTimeStamp';
  adFileTime         : Result := 'adFileTime';
  adDBFileTime       : Result := 'adDBFileTime';
  adLongVarChar      : Result := 'adLongVarChar';
  adLongVarWChar     : Result := 'adLongVarWChar';
  adBinary           : Result := 'adBinary';
  adVarBinary        : Result := 'adVarBinary';
  adLongVarBinary    : Result := 'adLongVarBinary';
  adGUID             : Result := 'adGUID';
  adEmpty            : Result := 'adEmpty';
  adError            : Result := 'adError';
  adArray            : Result := 'adArray';
  adChapter          : Result := 'adChapter';
  adIDispatch        : Result := 'adIDispatch';
  adIUnknown         : Result := 'adIUnknown';
  adPropVariant      : Result := 'adPropVariant';
  adUserDefined      : Result := 'adUserDefined';
  adVariant          : Result := 'adVariant';
 end;
 if HasFlags(FieldType,adArray) then begin
  if (Result='')
  then Result:='adArray'
  else Result:=Result+'Array';
 end;
end;

 // Converts an SQLDB native types into string related.
function TDbConverter.SqlDbFieldTypeToString(FieldType:TFieldType):LongString;
begin
 if UsesSystemDbFieldtypenames then
 if InRange(Ord(FieldType),Ord(Low(TFieldType)),Ord(High(TFieldType))) then begin
  Result:=Format('ft%s',[db.Fieldtypenames[FieldType]]);
  Exit;
 end;
 Result:='';
 case FieldType of
  ftUnknown       : Result:='ftUnknown';       // 0  Unknown data type
  ftString        : Result:='ftString';        // 1  String data value (ansistring)
  ftSmallint      : Result:='ftSmallint';      // 2  Small integer value(1 byte, signed)
  ftInteger       : Result:='ftInteger';       // 3  Regular integer value (4 bytes, signed)
  ftWord          : Result:='ftWord';          // 4  Word-sized value(2 bytes, unsigned)
  ftBoolean       : Result:='ftBoolean';       // 5  Boolean value
  ftFloat         : Result:='ftFloat';         // 6  Floating point value (double)
  ftCurrency      : Result:='ftCurrency';      // 7  Currency value (4 decimal points)
  ftBCD           : Result:='ftBCD';           // 8  Binary Coded Decimal value (DECIMAL and NUMERIC SQL types)
  ftDate          : Result:='ftDate';          // 9  Date value
  ftTime          : Result:='ftTime';          // 10 Time value
  ftDateTime      : Result:='ftDateTime';      // 11 Date/Time (timestamp) value
  ftBytes         : Result:='ftBytes';         // 12 Array of bytes value, fixed size (unytped)
  ftVarBytes      : Result:='ftVarBytes';      // 13 Array of bytes value, variable size (untyped)
  ftAutoInc       : Result:='ftAutoInc';       // 14 Auto-increment integer value (4 bytes)
  ftBlob          : Result:='ftBlob';          // 15 Binary data value (no type, no size)
  ftMemo          : Result:='ftMemo';          // 16 Binary text data (no size)
  ftGraphic       : Result:='ftGraphic';       // 17 Graphical data value (no size)
  ftFmtMemo       : Result:='ftFmtMemo';       // 18 Formatted memo data value (no size)
  ftParadoxOle    : Result:='ftParadoxOle';    // 19 Paradox OLE field data (no size)
  ftDBaseOle      : Result:='ftDBaseOle';      // 20 Paradox OLE field data
  ftTypedBinary   : Result:='ftTypedBinary';   // 21 Binary typed data (no size)
  ftCursor        : Result:='ftCursor';        // 22 Cursor data value (no size)
  ftFixedChar     : Result:='ftFixedChar';     // 23 Fixed character array (string)
  ftWideString    : Result:='ftWideString';    // 24 Widestring (2 bytes per character)
  ftLargeint      : Result:='ftLargeint';      // 25 Large integer value (8-byte)
  ftADT           : Result:='ftADT';           // 26 ADT value
  ftArray         : Result:='ftArray';         // 27 Array data
  ftReference     : Result:='ftReference';     // 28 Reference data
  ftDataSet       : Result:='ftDataSet';       // 29 Dataset data (blob)
  ftOraBlob       : Result:='ftOraBlob';       // 30 Oracle BLOB data
  ftOraClob       : Result:='ftOraClob';       // 31 Oracle CLOB data
  ftVariant       : Result:='ftVariant';       // 32 Variant data value
  ftInterface     : Result:='ftInterface';     // 33 interface data value
  ftIDispatch     : Result:='ftIDispatch';     // 34 Dispatch data value
  ftGuid          : Result:='ftGuid';          // 35 GUID data value
  ftTimeStamp     : Result:='ftTimeStamp';     // 36 Timestamp data value
  ftFMTBcd        : Result:='ftFMTBcd';        // 37 Formatted BCD (Binary Coded Decimal) value.
  ftFixedWideChar : Result:='ftFixedWideChar'; // 38 Fixed wide character date (2 bytes per character)
  ftWideMemo      : Result:='ftWideMemo';      // 39 Widestring memo data
 end;
end;

function TDbConverter.DbFieldTypeMap:TStringList;
var ad:Integer; ft:TFieldType; Ptr:Pointer; S:LongString;
begin
 if not Assigned(myDbFieldTypeMap) then begin
  myDbFieldTypeMap:=TStringList.Create;
  myDbFieldTypeMap.Duplicates:=dupIgnore;
  myDbFieldTypeMap.Sorted:=True;
  for ad:=Low(Byte) to High(Byte) do begin
   if ad in adScalarFieldTypes then begin
    Ptr:=PtrIntToPointer(ad);
    S:=AdoFieldTypeToString(ad);
    if (S<>'') then myDbFieldTypeMap.AddObject(S,Ptr);
    Ptr:=PtrIntToPointer(ad or adArray);
    S:=AdoFieldTypeToString(ad or adArray);
    if (S<>'') then myDbFieldTypeMap.AddObject(S,Ptr);
   end;
  end;
  myDbFieldTypeMap.AddObject('adArray',PtrIntToPointer(adArray));
  for ft:=Low(TFieldType) to High(TFieldType) do begin
   Ptr:=PtrIntToPointer(Ord(ft));
   S:=SqlDbFieldTypeToString(ft);
   if (S<>'') then myDbFieldTypeMap.AddObject(S,Ptr);
  end;
 end;
 Result:=myDbFieldTypeMap;
end;

function TDbConverter.StringToFieldTypeCode(const S:LongString; Def:Integer=-1):Integer;
var i:Integer;
begin
 Result:=Def;
 if (S<>'') then
 try
  i:=DbFieldTypeMap.IndexOf(S);
  if (i>=0) then Result:=PointerToPtrInt(DbFieldTypeMap.Objects[i]);
 except
  on E:Exception do BugReport(E,nil,'StringToDbFieldTypeCode');
 end;
end;

function TDbConverter.CodeToFieldType(Code:Integer; Def:TFieldType=ftUnknown):TFieldType;
begin
 if InRange(Code,Ord(Low(TFieldType)),Ord(High(TFieldType)))
 then Result:=TFieldType(Code)
 else Result:=Def;
end;

function TDbConverter.FieldTypeCodeToString(Code:Integer; Engine:Integer):LongString;
begin
 Result:='';
 case Engine of
  db_engine_code_ado:   Result:=AdoFieldTypeToString(Code);
  db_engine_code_sqldb: Result:=SqlDbFieldTypeToString(CodeToFieldType(Code));
 end;
end;

function TDbConverter.IsDateTimeFieldTypeCode(typ,Engine:Integer):Boolean;
const AdoCodes=[adDate,adDBDate,adDBTime,adDBTimeStamp,adFileTime,adDBFileTime];
const SqlDvCodes=[ftDate,ftTime,ftDateTime,ftTimeStamp];
begin
 case Engine of
  db_engine_code_ado:   Result:=(typ in AdoCodes);
  db_engine_code_sqldb: Result:=(CodeToFieldType(typ) in SqlDvCodes);
  else                  Result:=False;
 end;
end;

function TDbConverter.AdoDateTimeToMs(tm:Double; typ:Integer):Double;
begin
 case typ of
  adDate:        Result:=OleTimeToMs(tm);
  adDBDate:      Result:=OleTimeToMs(tm);
  adDBTime:      Result:=OleTimeToMs(tm);
  adDBTimeStamp: Result:=OleTimeToMs(tm);
  adFileTime:    Result:=FileTimeToMs(tm);
  adDBFileTime:  Result:=FileTimeToMs(tm);
  else           Result:=tm;
 end;
end;

function TDbConverter.SqlDbDateTimeToMs(tm:Double; typ:Integer):Double;
var ft:TFieldType;
begin
 ft:=CodeToFieldType(typ);
 case ft of
  ftDate:      Result:=OleTimeToMs(tm);
  ftTime:      Result:=OleTimeToMs(tm);
  ftDateTime:  Result:=OleTimeToMs(tm);
  ftTimeStamp: Result:=OleTimeToMs(tm);
  else         Result:=tm;
 end;
end;

function TDbConverter.DbDateTimeToMs(tm:Double; typ,Engine:Integer):Double;
begin
 case Engine of
  db_engine_code_ado:   Result:=AdoDateTimeToMs(tm,typ);
  db_engine_code_sqldb: Result:=SqlDbDateTimeToMs(tm,typ);
  else                  Result:=tm;
 end;
end;

///////////////////////////////////////
// Unit initialization and finalization
///////////////////////////////////////

procedure Init_crw_dbcon;
begin
 DbCon.Ok;
 DbCon.UsesSystemDbFieldtypenames:=True;
end;

procedure Free_crw_dbcon;
begin
 Kill(TheDbCon);
end;

initialization

 Init_crw_dbcon;

finalization

 Free_crw_dbcon;

end.

//////////////
// END OF FILE
//////////////

